/*
.wy-nav-content {
	max-width: 800px;
}
*/

.hll {
	background-color: #f2f8fb;
}

/*
.wy-nav-content {
	max-width: 100%;
}
*/

.wy-breadcrumbs-aside {
	display: none !important;
}

.wy-menu-vertical li.current a {
	color: #545454;
}

.toctree-l1 a, .toctree-l2 a, .toctree-l3 a {
	min-height: 32px;
}

/* Temporary fix for https://github.com/rtfd/sphinx_rtd_theme/issues/417 */

.rst-content .highlight>pre {
    line-height: 18px;
}

.rubric {
	/* font-size: 125%; */
	font-weight: 700;
    font-family: "Roboto Slab","ff-tisa-web-pro","Georgia",Arial,sans-serif;
}

em {
    font-style: normal;
}
